from __future__ import annotations

import os
from typing import cast, Collection, Optional, Tuple, Union

import torch


def _check_points(points: torch.Tensor) -> None:
    """Check points validity (type, shape)"""
    assert len(points.shape) == 2
    assert points.shape[0] > 0
    assert points.shape[1] in (2, 3)  # Only 2d and 3d supported
    assert points.dtype is torch.float32


class Track:
    """Track for a given particle

    A track is defined by an (non-unique) identifier, a starting frame and a succession of positions.

    Attributes:
        identifier (int): Identifier of the track (non-unique)
        start (int): Starting frame of the track
        points (torch.Tensor): Positions (i, j) of the particle (from starting frame to ending frame)
            Shape: (T, D), dtype: float32

    """

    _next_identifier = 0

    def __init__(self, start: int, points: torch.Tensor, identifier: Optional[int] = None) -> None:
        if identifier is None:
            self.identifier = Track._next_identifier
            Track._next_identifier += 1
        else:
            self.identifier = identifier

        _check_points(points)

        self.start = start
        self.points = points

    def __len__(self) -> int:
        return self.points.shape[0]

    def __getitem__(self, frame_id: int) -> torch.Tensor:
        """Return the position of tracked particle for the given frame_id

        Args:
            frame_id (int): id of the frame in the video

        Returns:
            torch.Tensor: Position (i, j) at the given frame (NaN if unknown)
                Shape: (D, ), dtype: float32

        """
        if self.start <= frame_id < self.start + len(self.points):
            return self.points[frame_id - self.start]

        return torch.full((self.points.shape[1],), torch.nan)

    def overlaps_with(self, other: Track, tolerance=0) -> bool:
        """Test if this track overlaps with another one in time.

        Args:
            other (Track): The other track
            tolerance (int): Time tolerance before detecting an overlap.
                Default: 0 (no tolerance)

        """
        return self.start < other.start + len(other) - tolerance and self.start + len(self) - tolerance > other.start

    @staticmethod
    def tensorize(tracks: Collection[Track], frame_range: Optional[Tuple[int, int]] = None) -> torch.Tensor:
        """Convert a collection of tracks into a tensor on a given frame_range

        Useful view of the data to speedup some mathematical operations

        Args:
            tracks (Collection[Track]): A collection of tracks (usually from the same video)
            frame_range (Optional[Tuple[int, int]]): Frame range (start included, end excluded)
                If None, the minimal range to hold all points is computed

        Returns:
            torch.Tensor: Tracks data in a single tensor
                Shape: (T, N, D), dtype: float32

        """
        # Find the spatial dimension (all tracks should share the same one)
        dim = next(iter(tracks)).points.shape[1]

        # Compute once start and end of each track
        starts = [track.start for track in tracks]
        ends = [track.start + len(track) for track in tracks]

        if frame_range is None:
            start = min(starts)
            end = max(ends)
        else:
            start, end = frame_range

        points = torch.full((end - start, len(tracks), dim), torch.nan)

        for i, (track_start, track_end, track) in enumerate(zip(starts, ends, tracks)):
            if track_end <= start or track_start >= end:
                continue

            track_end = min(end, track_end)

            points[max(0, track_start - start) : track_end - start, i, :] = track.points[
                max(0, start - track_start) : track_end - track_start
            ]

        return points

    @staticmethod
    def save(tracks: Collection[Track], path: Union[str, os.PathLike]) -> None:
        """Save a collection of tracks to path

        Format: pt (pytorch)

        .. code-block::

            {
                "offset": int
                "ids": Tensor (N, ), int64
                "points": Tensor (T, N, D), float32
            }

        Args:
            tracks (Collection[Track]): Tracks to save
            path (str | os.PathLike): Output path

        """
        ids = torch.tensor([track.identifier for track in tracks])
        offset = min(track.start for track in tracks)
        points = Track.tensorize(tracks)
        torch.save({"offset": offset, "ids": ids, "points": points}, path)

    @staticmethod
    def load(path: Union[str, os.PathLike]) -> Collection[Track]:
        """Load a collection of tracks from path

        Args:
            path (str | os.PathLike): Input path

        """
        data: dict = torch.load(path, map_location="cpu")
        offset: int = data.get("offset", 0)
        points: torch.Tensor = data["points"]
        ids: torch.Tensor = data["ids"]

        frames = torch.arange(points.shape[0])

        tracks = []

        for i, identifier in enumerate(ids.tolist()):
            track_points = points[:, i]
            defined = ~torch.isnan(track_points).all(dim=-1)
            track_points = track_points[defined]
            start = cast(int, frames[defined].min().item())
            tracks.append(Track(start + offset, track_points, identifier))

        return tracks



from typing import Callable, Collection, Dict, Iterable, List, Set

import numpy as np
import pylapy
import torch

import byotrack


Dist = Callable[[Iterable[np.ndarray], Collection[byotrack.Track]], np.ndarray]


class DistStitcher(byotrack.Refiner):
    """Track stitching using distance minimization

    Attributes:
        dist (Dist): Callable that compute the distance between each track
        eta (float): Soft threshold in LAP solving (See pylapy)
            Default: inf (No soft thresholding)
        lap_solver (LapSolver): Solver of the assignment problem

    """

    def __init__(self, dist: Dist, eta=float("inf")) -> None:
        super().__init__()
        self.dist = dist
        self.eta = eta
        self.lap_solver = pylapy.LapSolver()

    def run(self, video: Iterable[np.ndarray], tracks: Collection[byotrack.Track]) -> List[byotrack.Track]:
        dist = self.dist(video, tracks)
        links = self.lap_solver.solve(dist, self.eta)
        return self.merge(tracks, links)

    @staticmethod
    def merge(tracks: Collection[byotrack.Track], links: np.ndarray) -> List[byotrack.Track]:
        """Merge tracks following the given links

        Args:
            tracks (Collection[byotrack.Track]): Tracks to merge
            links (np.ndarray): Links between tracks. Each link is a pair of track indices
                Shape: (L, 2), dtype: int

        Returns:
            List[byotrack.Track]: Merged tracks. Note that in a merge track there is a probably a unknown gap
                between the two original tracks where the position of the merge track is set to nan by default
        """
        connected_components = _extract_connected_components(len(tracks), dict(links.tolist()))
        print(f"Merging {len(tracks)} tracks into {len(connected_components)} resulting tracks")

        track_list = list(tracks)
        merged_tracks = []
        dim = track_list[0].points.shape[1]

        for connected_component in connected_components:
            if len(connected_component) == 1:  # Do not duplicate the data just reuse the existing track
                merged_tracks.append(track_list[connected_component.pop()])
                continue

            # Merge multiple tracks

            start = min(track_list[i].start for i in connected_component)
            end = max(track_list[i].start + len(track_list[i]) for i in connected_component)

            points = torch.full((end - start, dim), torch.nan)

            for i in connected_component:
                points[track_list[i].start - start : track_list[i].start - start + len(track_list[i])] = track_list[
                    i
                ].points

            merged_tracks.append(byotrack.Track(start, points))

        return merged_tracks

    @staticmethod
    def skip_computation(
        tracks: Collection[byotrack.Track], max_overlap: int, max_dist: float, max_gap: int
    ) -> torch.Tensor:
        """Compute a boolean mask that indicate which distance should be skipped

        Based on simple rules, prevents the computation of most distances. Let i, j two tracks:

        1. i ends at most `max_gap` frames before j starts
        2. i ends at most `max_overlap` frames after j starts
        3. i last position is at most at `max_dist` from j first position

        Args:
            tracks (Collection[byotrack.Track]): Current set of tracks
            max_overlap (int): Cannot stitch tracks that overlap more than `max_overlap`
            max_dist (float): Cannot stich track i and track j if the last position of i and
                first position of j are farther than `max_dist` (ignored if max_dist <= 0)
            max_gap (int): Cannot stich track i and track j if i ended more
                than `max_gap` frame before j started (ignored if max_gap <= 0)

        Returns:
            torch.Tensor: Boolean tensor that indicates True when the dist computation should be skipped
        """
        starts = torch.tensor([track.start for track in tracks])
        ends = torch.tensor([track.start + len(track) for track in tracks])

        first_pos = torch.concatenate([track.points[:1] for track in tracks])
        last_pos = torch.concatenate([track.points[-1:] for track in tracks])

        skip = starts[:, None] >= starts[None, :]  # Ensure full asymmetry
        skip |= ends[:, None] > starts[None, :] + max_overlap

        if max_gap > 0:
            skip |= ends[:, None] + max_gap < starts[None, :]

        if max_dist > 0:
            skip |= (last_pos[:, None, :] - first_pos[None, :, :]).pow(2).sum(dim=-1) > max_dist**2

        return skip

    @staticmethod
    def normalize(dist: np.ndarray) -> np.ndarray:
        """Normalize a distance matrix into [0, 1]

        Args:
            dist (np.ndarray): Distance matrix to normalize
                Shape: (N, N), dtype: float32

        Returns:
            np.ndarray: Normalized distance matrix
                Shape: (N, N), dtype: float32
        """
        return (dist - dist.min()) / (dist[dist != torch.inf].max() - dist.min())


def _extract_connected_components(n: int, edges: Dict[int, int]) -> List[Set[int]]:
    """Extract the connected components in a linear directed graph

    Assume that each node has at most one child.

    Args:
        n (int): Number of nodes (from 0 to n - 1)
        edges (Dict[int, int]): Edges i -> j

    Returns:
        List[Set[int]]: List of connected components (each is a set of id)
    """
    edges = edges.copy()
    seen_nodes = np.full(n, False)
    start_nodes_to_cc_id: Dict[int, int] = {}

    connected_components: List[Set[int]] = []

    for node in range(n):
        if seen_nodes[node]:  # If seen, no need to go through it again
            continue

        seen_nodes[node] = True
        connected_component = {node}

        start_node = node

        while node in edges:  # Pop each edges from the start node a follow them
            node = edges.pop(node)
            seen_nodes[node] = True
            connected_component.add(node)

        # As nodes are not sorted we could find at the end an already handled start_node
        # Let's then just merge the connected components
        if node in start_nodes_to_cc_id:
            cc_id = start_nodes_to_cc_id[node]
            connected_components[cc_id].update(connected_component)
        else:  # New connected components
            cc_id = len(connected_components)
            connected_components.append(connected_component)

        start_nodes_to_cc_id[start_node] = cc_id

    return connected_components



import enum
import os
import pathlib
from typing import Collection, Iterable, Optional, Union

import numpy as np

import byotrack
from byotrack import icy
from byotrack.api.parameters import ParameterEnum


PROTOCOL = pathlib.Path(__file__).parent / "emht_protocol.xml"


class IcyEMHTLinker(byotrack.Linker):  # pylint: disable=too-few-public-methods
    """Run EMHT [4] from Icy [1]

    This code is only a wrapper arounds Icy implementation as EMHT is painful to implement.

    About EMHT:
    It is a probabilistic tracking that uses statistical motion model on particles. It uses multiple
    kalman filters for each particle allowing a particle to have several mode of motions. It also keeps
    multiple hypothesis of tracking at each frame so that a final detections linking decision is made
    after seeing several frames in the past and future of these detections.

    Here we rely on the handmade protocol "emht_protocol" that expects detections as a valid rois file
    for icy and some hyperparameters.

    The workflow is:

    1. Detections to rois in Icy format
    2. Run the Icy protocol
    3. [In ICY] Read rois, estimate emht parameters, run emht, export tracks to xml
    4. Read Icy tracks and return

    Note:
        This implementation requires Icy to be installed (https://icy.bioimageanalysis.org/download/)

    Attributes:
        runner (byotrack.icy.IcyRunner): Icy runner
        motion (Motion): Prior on the underlying motion model (Brownian vs Directed vs Switching)
            Given to the Icy block that estimates EMHT parameters.
            Default: Motion.BROWNIAN

    ## XXX: Should we add update_cov (bool): Whether to update covariances in Kalman filters

    """

    class Motion(enum.Enum):
        """Different motion models:

        Brownian: Random gaussian displacement at each time
        Directed: Random gaussian noise around a directed trajectory
        Multi: Switch randomly between Brownian and Directed

        """

        BROWNIAN = (0, 0)
        DIRECTED = (1, 0)
        MULTI = (0, 1)

    parameters = {"motion": ParameterEnum({Motion.BROWNIAN, Motion.DIRECTED, Motion.MULTI})}
    rois_file = "_tmp_rois.xml"
    tracks_file = "_tmp_tracks.xml"

    def __init__(self, icy_path: Optional[Union[str, os.PathLike]] = None) -> None:
        """Constructor

        Args:
            icy_path (str | os.PathLike): Path to the icy jar (Icy is called with java -jar <icy_jar>)
                If not given, icy is searched in the PATH
        """
        super().__init__()
        self.runner = icy.IcyRunner(icy_path)
        self.motion = IcyEMHTLinker.Motion.BROWNIAN

    def run(
        self, video: Iterable[np.ndarray], detections_sequence: Collection[byotrack.Detections]
    ) -> Collection[byotrack.Track]:
        try:
            icy.save_detections(detections_sequence, self.rois_file)
            self._run_icy()
            # Sort tracks by starting time and then position (They seem to be randomly sorted otherwise and in addition
            # with EMC2 torch-tps approximate propagation it yields undeterministic behaviors)
            return sorted(
                icy.load_tracks(self.tracks_file), key=lambda track: (track.start, track.points[0].sum().item())
            )
        finally:
            if os.path.exists(self.rois_file):
                os.remove(self.rois_file)
            if os.path.exists(self.tracks_file):
                os.remove(self.tracks_file)

    def _run_icy(self):
        """Run the icy process"""
        directed, multi = self.motion.value

        self.runner.run(
            PROTOCOL,
            rois=os.path.abspath(self.rois_file),
            tracks=os.path.abspath(self.tracks_file),
            directed=directed,
            multi=multi,
        )

        # Has to check because icy do not return any non-zero return code
        if not os.path.exists(self.tracks_file):
            raise RuntimeError(
                """No track found from Icy.

            This probably results from a failure in Icy software. Please look at the Java Exception displayed by Icy.

            Note that Icy may fail the first time you launch this code because Icy updates the required protocols
            and is unable to continue to run after the update.
            """
            )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection, Iterable

import numpy as np

import byotrack  # pylint: disable=cyclic-import
from .parameters import ParametrizedObjectMixin


class Refiner(ABC, ParametrizedObjectMixin):  # pylint: disable=too-few-public-methods
    """Base class for tracks refiners

    Improve a set of tracks

    Each refiner can define a set of parameters (See `ParametrizedObjectMixin`)

    """

    @abstractmethod
    def run(self, video: Iterable[np.ndarray], tracks: Collection[byotrack.Track]) -> Collection[byotrack.Track]:
        """Run the refiner on a whole video

        Args:
            video (Iterable[np.ndarray]): Sequence of frames (video)
                Each array is expected to have a shape (H, W, C)
            tracks (Collection[byotrack.Track]): Tracks of particles

        Returns:
            Collection[byotrack.Track]: Refined tracks of particles

        """



"""Video package

Provide a Video class that reads video from disk. It implements slicing and iterable protocols.
Also provide default transforms to scale video intensities and select a given channel.
"""

from .video import Video, VideoTransformConfig
